/*
  -------------------- file ntup.c
  Created 18 Nov 93 by Andrew Shapira, Rensselaer Polytechnic Institute.

	Copyright (C) 1993 Andrew Shapira.

	This file is part of the Ntuplib library.

	Ntuplib is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	Ntuplib is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with the Ntuplib Library; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
  Terminology, conventions:
    pi     The area that is the intersection of the characters' bounding
    rectangles.  Tuple coordinates are selected from this region.  Pi
    contains no information about white/black pixels; it is simply a
    coordinate range from which tuple coordinates are selected.

    pch     positive character index (0 .. npos)
    nch     negative character index (0 .. nneg)
*/

#include <limits.h>

#include "our_common.h"
#include "../libinclude/ntup.h"
#include "../libinclude/tuputil.h"

#include "prog/matrix.h"
#include "prog/memalloc.h"

/*  Display message associated with status field of ntup_from structure.
    Does not display trailing newline.
*/
void
ntup_print_error(
FILE *f,
const struct from_ntup_struct *from
)
{
  const int status = from->status;

  switch (status) {
  case NTUP_OK:
    fprintf(f, "Completed without error.");
    break;
  case NTUP_NO_POS_GIVEN:
    fprintf(f, "No positive characters.");
    break;
  case NTUP_NO_NEG_GIVEN:
    fprintf(f, "No negative characters.");
    break;
  case NTUP_INVALID_N:
    fprintf(f, "Value of n is too small or too large.");
    break;
  case NTUP_INVALID_TABU_ITERATIONS:
    fprintf(f, "Invalid value of tabu_max_iterations.");
    break;
  case NTUP_INVALID_PQKIND:
    fprintf(f, "Invalid search type (pq_kind field).");
    break;
  case NTUP_INVALID_PORQ:
    fprintf(f, "Invalid value or p or q.");
    break;
  case NTUP_PI_TOO_SMALL:
    fprintf(f,
      "Pi is too small (n is too big for the specified pi dimensions).");
    break;
  case NTUP_INVALID_BACK_MAX_NODES:
    fprintf(f, "Invalid value of back_max_nodes.");
    break;
  case NTUP_INVALID_UPDATE_INTERVAL:
    fprintf(f, "Update interval must be %d or more", UPDATE_MIN);
    break;
  case NTUP_INVALID_TUPLE_KIND:
    fprintf(f, "Invalid tuple kind (must be %d, %d, or %d).",
      TUPLE_BLACKONLY, TUPLE_WHITEONLY, TUPLE_BLACKWHITE);
    break;
  case NTUP_INVALID_BACK2_TIMESLICE_NODES:
    fprintf(f, "Invalid limit specified for timeslice nodes (must be 0 or more).");
    break;
  case NTUP_INVALID_BACK_ACTION:
    fprintf(f, "Allowed values of back_action field are %d and %d.",
      BACK_ACTION_COUNT, BACK_ACTION_ONE);
    break;
  case NTUP_INVALID_BACK_PRINT:
    fprintf(f, "Value of back_print field must be in range %d .. %d.",
      0, BACK_PRINT_MAX);
      break;
  case NTUP_INVALID_BACK2_N:
    fprintf(f, "With back2 search, n must be 3 or more.");
    break;
  case NTUP_INVALID_BACK2_RESTRICT:
    fprintf(f, "Value of back2_restrict_width must 0 or more.");
    break;
  case NTUP_INVALID_WIDE1_WIDTH:
    fprintf(f, "Value of wide1_width must be 0 or more.");
    break;
  case NTUP_INVALID_WIDE1_N:
    fprintf(f, "With wide1 search, n must be 3 or more.");
    break;
  case NTUP_INVALID_PIXEL_VALUE:
    fprintf(f, "Input pixel value out of range.");
    break;
  case NTUP_INVALID_POS_BG:
    fprintf(f, "Background for positive class must be in range %d .. %d.\n",
      0, PSEUDOCOLOR_MAX);
    break;
  case NTUP_INVALID_NEG_BG:
    fprintf(f, "Background for negative class must be in range %d .. %d.\n",
      0, PSEUDOCOLOR_MAX);
    break;
  case NTUP_INVALID_PI_DIMENSIONS:
    fprintf(f, "Dimensions of pi must be positive.\n");
    break;
  default:
    fprintf(f, "Unknown ntup_find error code.");
  }
}

/*  Free space associated with the fields of a from_ntup_struct.  The
    fields of the structure are assumed to have been allocated using
    malloc().
*/
void
ntup_from_free(
struct from_ntup_struct *from
)
{
  tuplelist_free(from->tuples, from->tuples_allocated, xfree);  
}


/*  Display an ntuple.

    TODO - much of this is cloned in the old ocr program.  Make that program
    use this function.
*/
void
ntup_print(
FILE *f,
const struct pix_struct *pixels,
const int n,
const int print_format              /* 0=short format, 1=long format */
)
{

  int   j;
  int   min_row,max_row,min_col,max_col;

  if (n == 0)  {  fprintf(f,"(empty tuple)\n");  return;  }

  min_row = min_col = INT_MAX;
  max_row = max_col = INT_MIN;
  fprintf(f, "Abs:  ");
  for (j = 0; j < n; j++) {
    const int row = pixels[j].row;
    const int col = pixels[j].col;

    if (j > 0)  fprintf(f, " ");
    fprintf(f,"%c(%2d,%2d)", pixels[j].color ? 'B' : 'W', row, col);

    if (row < min_row) min_row = row;
    if (row > max_row) max_row = row;
    if (col < min_col) min_col = col;
    if (col > max_col) max_col = col;
  }
  fprintf(f,"\n");

  fprintf(f, "Rel:  ");
  for (j = 0; j < n; j++) {
    const int row = pixels[j].row;
    const int col = pixels[j].col;

    if (j > 0)  fprintf(f, " ");
    fprintf(f,"%c(%2d,%2d)",
      pixels[j].color ? 'B' : 'W', row-min_row, col-min_col);
  }
  fprintf(f,"\n");

  {
    const int rows = max_row-min_row+1;
    const int cols = max_col-min_col+1;
    int   r,c;
    unsigned char **m;
    const unsigned char EMPTY = 255;

    bsbegin();
    m = mat_uchar_alloc2(rows,cols,bsalloc);
    mat_uchar_set2(m,rows,cols,EMPTY);
    for (j = 0; j < n; j++) {
      m [pixels[j].row-min_row] [pixels[j].col-min_col] = pixels[j].color;
    }
    for (r = 0; r < rows; r++) {
      for (c = 0; c < cols; c++) {
        if (print_format && (c > 0))  fprintf(f," ");
        if (m[r][c] != EMPTY)         fprintf(f,"%c", m[r][c] ? 'B' : 'W');
        else                          fprintf(f,".");
      }
      if (print_format) fprintf(f,"\n");
      else              fprintf(f, " ; ");
    }
    if (! print_format) fprintf(f,"\n");
    bsend();
  }
}

void
ntup_trans_make_rectangular(
struct character_struct *const h,   /* we fill in h's translation info */
const int mindr,
const int maxdr,
const int mindc,
const int maxdc,
void  *(*allocfunc) P((const int))
)
{
  const int ntrans = (maxdr-mindr+1) * (maxdc-mindc+1);
  int d, dr, dc;

  ASSERT(mindr <= maxdr);
  ASSERT(mindc <= maxdc);
  h->ntrans = ntrans;
  h->trans_list = (*allocfunc) (ntrans * sizeof(h->trans_list[0]));
  d = 0;
  for (dr = mindr; dr <= maxdr; dr++) {
    for (dc = mindc; dc <= maxdc; dc++) {
      h->trans_list[d].dr = dr;
      h->trans_list[d].dc = dc;
      d++;
    }
  }
  ASSERT(d == ntrans);
}
