/*
  -------------------- file run.c
  Created 3 May 94 by Andrew Shapira, Rensselaer Polytechnic Institute.
  The main purpose of this file is to have a function that calls the
  search procedures with the correct values of p and q, and keeps track
  of solutions as needed.

	Copyright (C) 1994 Andrew Shapira.

	This file is part of the Ntuplib library.

	Ntuplib is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	Ntuplib is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with the Ntuplib Library; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* #define DEBUG */

#include "our_common.h"
#include "../libinclude/ntup.h"
#include "../libinclude/tuputil.h"
#include "charutil.h"
#include "run.h"
#include "pi.h"
#include "tmask.h"
#include "prog/matrix.h"
#include "prog/memalloc.h"


/*  Given a list of tuples, generate a new list L so that no tuple in L
    is dominated by any other tuple of L.  Note: if some tuples
    in the original list have the same p,q values, then at most one of
    these tuples will be in L.

    No assumption is made about how the original tuple list was generated.
    They may have been generated by PQ_QTHENP, or by some other method.

    This function is stable, i.e., the order of tuples that are kept
    is preserved.

    Note: the table for existing tuples is staircase shaped.  However,
    the (heuristic) search may not find tuples in a staircase shape.
*/
static void
remove_dominated(
struct tuple_struct *tuples,
const int len_in,                     /* number of tuples originally in list */
int *len_ret                    /* we set to new length of list */
)
{
  int i,j,len;

  for (i = 0, len = len_in  ;  i < len  ; ) {
    const int ip = tuples[i].p;
    const int iq = tuples[i].q;
    int idominated;                 /* was i found to be dominated? */

    /*  Compare i to all other tuples; remove i if it is dominated by any.
    */
    for (j = 0, idominated = 0  ;  j < len  ;  j++) {
      if (i != j) {
        const int jp = tuples[j].p;
        const int jq = tuples[j].q;

        if ((ip <= jp) && (iq >= jq))  {  idominated = 1;  break;  }
      }
    }

    if (idominated) {
      for (j = i+1; j < len; j++)
        tuple_copy(& tuples[j-1], & tuples[j]);
      len--;
    }
    else
      i++;
  }

  *len_ret = len;
}

static void
report_results(
FILE * const f,
const struct from_ntup_struct *const from
)
{
  int i;

  fprintf(f, "---------- Begin report of generator results.\n");
  fprintf(f, "status=%d:  ", from->status);
  ntup_print_error(f,from);
  fprintf(f, "\n");
  fprintf(f, "tuples_allocated=%d  tuples_found=%d\n",
    from->tuples_allocated, from->tuples_found);
  if (from->tuples_found > 0) {
    for (i = 0; i < from->tuples_found; i++) {
      const struct tuple_struct * const tup = & from->tuples[i];
      fprintf(f, "-- Tuple %d    n=%d   (p,q)=(%d,%d)   pixels:\n",
        i, tup->n, tup->p, tup->q);
      ntup_print(f, tup->pixels, tup->n, 1);      /* long format */
    }
  }
  fprintf(f, "---------- End report of generator results.\n");
}

void
report_back(
FILE * const f,
const int     back_action,
const int     back_print
)
{
  fprintf(f,"back_action: ");
  if (back_action == BACK_ACTION_COUNT)
    fprintf(f, "count all solutions");
  else if (back_action == BACK_ACTION_ONE)
    fprintf(f, "quit after one solution");
  else
    fprintf(f, "ERROR");
  fprintf(f, "  ");
  
  fprintf(f,"back_print: ");
  if (back_print == BACK_PRINT_NONE)
    fprintf(f, "none");
  else if (back_print == BACK_PRINT_FIRST)
    fprintf(f, "first solution found");
  else if (back_print == BACK_PRINT_ALL)
    fprintf(f, "all solutions found");
  else if (back_print == BACK_PRINT_CANON_ONLY)
    fprintf(f, "all canonical solutions found");
  else
    fprintf(f, "ERROR");
  fprintf(f, "\n");
}

static int
coverpi(
const struct character_struct *h,
const int pi_rows,
const int pi_cols
)
{
  int mindr,maxdr,mindc,maxdc;

  if (! trans_form_rectangle(h,&mindr,&maxdr,&mindc,&maxdc))
    return(0);
  if (mindr != 1-pi_rows)
    return(0);
  if (maxdr != h->nrows-1)
    return(0);
  if (mindc != 1-pi_cols)
    return(0);
  if (maxdc != h->ncols-1)
    return(0);
  return(1);
}

static void
report_charclass(
FILE *const f,
const struct charclass_struct * const class,
const int pi_rows,
const int pi_cols,
const char * const str
)
{
  int i;
  int tstandard;    /* translations exactly "cover" each character with pi */

  fprintf(f,"----- %d %s character(s)", class->nchars, str);
  if (class->bg == WHITE)
    fprintf(f, ", white background");
  else if (class->bg == BLACK)
    fprintf(f, ", black background");
  else if (class->bg == PSEUDOCOLOR_NOMATCH)
    fprintf(f, ", no-match background");
  else if (class->bg == PSEUDOCOLOR_MATCH)
    fprintf(f, ", match background");
  else
    fprintf(f, ", ERROR background");

  for (i = 0, tstandard = 1; i < class->nchars; i++) {
    if (! coverpi(& class->chars[i],pi_rows,pi_cols))
      {  tstandard = 0;  break;  }
  }
  if (tstandard)
    fprintf(f, ", standard translations");
  fprintf(f, ".\n");

  for (i = 0; i < class->nchars; i++) {
    fprintf(f,"-- pixels of %s character %d:\n", str, i);
    ntup_print_char_pixels(f, & class->chars[i]);
    if (! tstandard) {
      fprintf(f,"-- translations for %s character %d:\n", str, i);
      print_char_trans(f, & class->chars[i]);
    }
  }
}

void
report_input(
const struct to_ntup_struct *const to
)
{
  FILE * const f = to->f;

  fprintf(f,"n=%d", to->n);

  fprintf(f, "   tuple_kind=%d ", to->tuple_kind);
  if (to->tuple_kind == TUPLE_BLACKONLY)
    fprintf(f,"(black only)");
  else if (to->tuple_kind == TUPLE_WHITEONLY)
    fprintf(f,"(white only)");
  else if (to->tuple_kind == TUPLE_BLACKWHITE)
    fprintf(f,"(black and white)");
  else
    fprintf(f,"(ERROR)");

  fprintf(f, "   pi: %d by %d\n", to->pi_rows, to->pi_cols);

  fprintf(f,"   pq_kind=%d", to->pq_kind);
  if (to->pq_kind == PQ_QTHENP)
    fprintf(f," (q then p)");
  else if (to->pq_kind == PQ_SPECIFIC)
    fprintf(f," (p,q)=(%d,%d)   ",to->p, to->q);
  else if (to->pq_kind == PQ_QTHENP_NOMAXP)
    fprintf(f," (q then p, do not maximize p internally)");
  else
    fprintf(f," (ERROR)");

  fprintf(f, "\n");

  fprintf(f,"update_interval=%d   ioflags=%08xx  ",
    to->update_interval, to->ioflags);
  if (to->f)
    fprintf(f, "f=non-NULL\n");
  else
    fprintf(f, "f=NULL\n");

  fprintf(f,"(Random number generator state=unknown.)\n");

  report_charclass(f,& to->pos,to->pi_rows,to->pi_cols,"positive");
  report_charclass(f,& to->neg,to->pi_rows,to->pi_cols,"negative");
}

/*
  This function should free all memory it allocates if it sets
  the status field to anything other than NTUP_OK.
*/
void
ntup_run(
const struct to_ntup_struct *to,
const int canon_only, 
  /* does work function generate only canonical n-tuples? */
struct from_ntup_struct *from,
void (*workfunc) P ((WORKFUNC_ARGS))    /* function we call to do the work */
)
{
  const int n = to->n;
  struct tuple_struct prev_solution[1];
  struct tuple_struct solution[1];
  struct coord_struct pi[1];
  struct tmask_struct pos_tmask[1];
  struct tmask_struct neg_tmask[1];
  int nfound;        /* number of tuples we put in return list */

  bsbegin();

  nfound = 0;                     /* haven't found any tuples yet */
  from->status = NTUP_OK;
  from->tuples = tuplelist_make(n,n,xmalloc); /* space for n n-tuples */
  from->tuples_allocated = n;

  tuple_alloc_fields(solution, n, bsalloc);
  tuple_alloc_fields(prev_solution, n, bsalloc);

  /* -------------------------------- Calculate invariants.
  */
  pi_make_standard(to->pi_rows, to->pi_cols, pi, bsalloc);
  if (pi->nids < n)
    {  from->status = NTUP_PI_TOO_SMALL;  goto done;  }

  tmask_make(to->pos.chars,to->pos.nchars,pi,to->pos.bg,pos_tmask,bsalloc);
  tmask_make(to->neg.chars,to->neg.nchars,pi,to->neg.bg,neg_tmask,bsalloc);

  D(printf("positive tmask\n");)
  D(debug_tmask(pos_tmask);)
  D(printf("negative tmask\n");)
  D(debug_tmask(neg_tmask);)

  /* -------------------------------- Carry out search.
  */

  switch (to->pq_kind) {
    /* -------------------------------- search a single specified (p,q)
    */
    case PQ_SPECIFIC: {
      int status;         /* return from work function */

      (*workfunc) (to->p,to->q,0,NULL,pi,pos_tmask,neg_tmask,solution,&status);
      if ((status==WORKFUNC_P) || (status==WORKFUNC_PMAX)) {
        tuple_copy(& from->tuples[0], solution);
        nfound = 1;
      }
      break;
    }                     /* case */


    /*  -------------------------------- search over all p,q.
        For each q: (search increasing p until failure).  We start at q=1
        and increase q, starting at the previous p and increasing p as
        possible.  We use the fact that if a tuple T satisfies the (p,q)
        constraint, then it satisfies (p,q+1).  The latest found tuple
        is always the starting tuple for the next search.

        The "work function" - the function that finds n-tuples - may
        be one of various forms:
          (a) a function that searches for given p,q
          (b) a function that, for a given q, finds the largest p
    */
    case PQ_QTHENP:         /* maxp */
    case PQ_QTHENP_NOMAXP:  /* no maxp */
      {
      int p,q;              /* what we are trying to find */
      int status;           /* return from work function */
      int test_p;           /* do a p test next time through loop? */

      for (test_p = 0, p = q = 1; q <= n-1; ) {
#ifdef DEBUG
printf("ntup_run: top of loop.  (p=%d,q=%d)\n", p, q);
if (p>1) {
  printf("ntup_run: prev_solution:\n");
  ntup_print(stdout, prev_solution->pixels, prev_solution->n, 1);
    /* 1 = use long format */
}
#endif /*DEBUG*/
        if (test_p && tmask_pos_satisfied
                      (prev_solution->pixels,prev_solution->n,p,pos_tmask)) {
D(printf("ntup_run: previous tuple satisfies p=%d\n", p);)
          prev_solution->p = p;   /* satisfies p now */
          p++;                    /* leave test_p set */
        }
        else {
D(printf("ntup_run: previous tuple does not satisfy p=%d; calling work function\n", p);)
          (*workfunc) (p,q,to->pq_kind == PQ_QTHENP,(p>1) ? prev_solution:NULL,
            pi,pos_tmask,neg_tmask,solution,&status);
D(printf("ntup_run: back from work.  status=%d\n", status);)
          ASSERT((status==WORKFUNC_NOTFOUND) || (status==WORKFUNC_NONEEXIST) ||
                  ((solution->p >= p) && (solution->q >= q)));
          switch (status) {
            case WORKFUNC_NOTFOUND:
            case WORKFUNC_NONEEXIST:
              if (p>1) {
                ASSERT(nfound < n);
                tuple_copy(& from->tuples[nfound++], prev_solution);
              }
              q++;
              test_p = 0;
              /* (leave p unchanged) */
              break;
            case WORKFUNC_P:
              tuple_copy(prev_solution, solution);
              p++;
              test_p = 1;
              break;
            case WORKFUNC_PMAX:
              ASSERT(nfound < n);
              tuple_copy(& from->tuples[nfound++], solution);
              tuple_copy(prev_solution, solution);
              p = solution->p + 1;
              q++;
              test_p = 0;
              break;
            default:
              fatal("ntup_run: work function returned invalid status.");
          }                 /* switch */
        }                   /* else */
      }                     /* for p */
      break;
    }                       /* PQ_QTHENP and PQ_QTHENP_NOMAXP */

    default:
      from->status = NTUP_INVALID_PQKIND;

  }     /* switch */

  if (from->status == NTUP_OK)
    remove_dominated(from->tuples,nfound,&nfound);
#ifdef ASSERTIONS_ON
  {
    int i;
    for (i = 0; i < nfound; i++) {
      const struct tuple_struct * const T = & from->tuples[i];
      if (tmask_pos_satisfied(T->pixels,T->n,T->p,pos_tmask) &&
          tmask_neg_satisfied(T->pixels,T->n,T->q,neg_tmask)) {}
      else {
        ntup_print(stdout, T->pixels, T->n, 1);
        ASSERT("invalid solution presented" == NULL);
      }
    }
  }
#endif  /*ASSERTIONS_ON*/

done:
  from->tuples_found = nfound;
  bsend();

  if (to->ioflags & IOF_REPORT_RESULTS)
    report_results(to->f, from);
}

/*  Note: do not use bit manipulation functions here, because we
    have not yet called bit_global_init().
*/
void
ntup_check_input(
const struct to_ntup_struct *to,
int *status_ret
)
{
  if (to->n < 2)
    {  *status_ret = NTUP_INVALID_N;  return;  }
  if (to->pq_kind == PQ_SPECIFIC) {
    if ((to->p < 1 ) || (to->q < 1) || (to->q > to->n))
      {  *status_ret = NTUP_INVALID_PORQ;  return;  }
  }
  if ((to->tuple_kind != TUPLE_BLACKONLY) &&
      (to->tuple_kind != TUPLE_WHITEONLY) &&
      (to->tuple_kind != TUPLE_BLACKWHITE))
    { *status_ret = NTUP_INVALID_TUPLE_KIND;  return;  }
  if (to->update_interval < UPDATE_MIN)
    {  *status_ret = NTUP_INVALID_UPDATE_INTERVAL;  return;  }
  if (to->pos.nchars < 1)
    {  *status_ret = NTUP_NO_POS_GIVEN;  return;  }
  if (to->neg.nchars < 1)
    {  *status_ret = NTUP_NO_NEG_GIVEN;  return;  }
  {
    int min,max,i;
    for (i = 0; i < to->pos.nchars; i++) {
      mat_uchar_minmax2( to->pos.chars[i].a,
        to->pos.chars[i].nrows,to->pos.chars[i].ncols,&min,&max);
      if ((min < PSEUDOCOLOR_MIN) || (max > PSEUDOCOLOR_MAX))
        {  *status_ret = NTUP_INVALID_PIXEL_VALUE;  return;  }

    }
    for (i = 0; i < to->neg.nchars; i++) {
      mat_uchar_minmax2(to->neg.chars[i].a,
        to->neg.chars[i].nrows,to->neg.chars[i].ncols,&min,&max);
      if ((min < PSEUDOCOLOR_MIN) || (max > PSEUDOCOLOR_MAX))
        {  *status_ret = NTUP_INVALID_PIXEL_VALUE;  return;  }
    }
  }
  if ((to->pos.bg < 0) || (to->pos.bg > PSEUDOCOLOR_MAX))
    {  *status_ret = NTUP_INVALID_POS_BG;  return;  }
  if ((to->neg.bg < 0) || (to->neg.bg > PSEUDOCOLOR_MAX))
    {  *status_ret = NTUP_INVALID_NEG_BG;  return;  }
  if ((to->pi_rows <= 0) || (to->pi_cols <= 0))
    {  *status_ret = NTUP_INVALID_PI_DIMENSIONS;  return;  }

  *status_ret = NTUP_OK;
}

/*  Check the back_xx fields.  Leaves *status_ret unchanged if no problems
    are detected.

    Note: do not use bit manipulation functions here, because we
    have not yet called bit_global_init().
*/
void
ntup_check_input_back(
const int back_action,
const int back_print,
const double back_max_nodes,
int * const status_ret
)
{
  if ((back_action != BACK_ACTION_COUNT) && (back_action != BACK_ACTION_ONE))
    {  *status_ret = NTUP_INVALID_BACK_ACTION;  return;  }
  if ((back_print < 0) || (back_print > BACK_PRINT_MAX))
    {  *status_ret = NTUP_INVALID_BACK_PRINT;  return;  }
  if (back_max_nodes < 0)
    {  *status_ret = NTUP_INVALID_BACK_MAX_NODES;  return;  }
}

void
lev_nodes_print(
FILE *const f,
const double * const lev_nodes,
const int n,
const int p,
const int q
)
{
  int i;

  fprintf(f, "Number of nodes at each level (p=%d,q=%d):\n",p,q);
  for (i = 0; i < n; i++)
    fprintf(f,"%8d", i);
  fprintf(f, "\n");
  for (i = 0; i < n; i++)
    fprintf(f,"%8g", lev_nodes[i]);
  fprintf(f, "\n");
}
