/*
  -------------------- file bitarray.h
  Created 30 Oct 93 by Andrew Shapira, Rensselaer Polytechnic Institute.

	Copyright (C) 1993 Andrew Shapira.

	This file is part of the Ntuplib library.

	Ntuplib is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	Ntuplib is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with the Ntuplib Library; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
  ////////////////////////////////////////////////////////////////////////////
  ///// See the comment in bitword.h about "BITWORD" and word sizes.
  ///// See comment(s) in bitarray.c.
  ////////////////////////////////////////////////////////////////////////////

  ////////////////////////////////////////////////////////////////////////////
  ///// USERS MUST CALL bit_global_init() before calling these routines.
  ////////////////////////////////////////////////////////////////////////////
*/

#ifndef _301093_180645
#define _301093_180645

#include "bitword.h"

/*  The bit positions in a bit array of length n are numbered 0 .. n-1.
*/

/* MACHINE DEPENDENT */
#define BIT_POW2                5     /* 5 for 2^5 = 32 */
#define BIT_RMASK_POW2          (0x0000001f)
#define BITS_PER_WORD           32    /* 2^BIT_POW2 */


/* number of "BITWORD"s used by a nn-bit array; use only for nn >= 0 */
#define BIT_WORDS(nn)       ( ((nn) + BITS_PER_WORD-1) >> BIT_POW2 )

/* word that bit bb is in in a bit array; use only for bb >= 0. */
#define BIT_WORD_INDEX(bb)  ( (bb) >> BIT_POW2 )
/* index of bit bb in word BIT_WORD_INDEX(bb); use only for bb >= 0 */
#define BIT_IN_WORD(bb)     ( (bb) & BIT_RMASK_POW2 )

/* mask for active bits in last word in bit array of length nn */
/* ----------- USE ONLY FOR nn > 0. ---------- */
/* COULDDO: it may be possible to make this faster and less ugly */
#define BIT_LASTMASK(nn)    \
    ( bitword_rmask [1+(((nn)+BITS_PER_WORD-1) & BIT_RMASK_POW2)] )


/* set, reset bit "bb" of array "aa" of BITWORDs; use only for bb >= 0.
   aa is BITWORD *; bb integer giving bit number.
*/
#define BIT_SET(aa,bb) \
          ( BITWORD_SET(aa[BIT_WORD_INDEX(bb)], BIT_IN_WORD(bb)) )
#define BIT_RESET(aa,bb) \
          ( BITWORD_RESET(aa[BIT_WORD_INDEX(bb)], BIT_IN_WORD(bb)) )
/* return zero if bb is not set in aa, else return nonzero */
#define IS_BIT_SET(aa,bb) \
          ( BITWORD_IS_BIT_SET((aa)[BIT_WORD_INDEX(bb)], BIT_IN_WORD(bb)) )
/* return value (0 or 1) of bit bb in aa */
#define BIT_VAL(aa,bb)    ( IS_BIT_SET((aa),(bb)) != 0 )


/*  Set up a loop that examines positions of a bit array that are 1.
      bb = bit array
      nn = length of bit array
      pp = an int variable that gives position of current 1 bit
      ww = an int variable for this macro's use; gives word of pp
    bb is left unchanged by the loop.

  This macro must be used as follows (with braces uglily unbalanced as shown):
    FOR_ONE_BITS(bb,nn,pp,ww)
      user code
    }

  The following is not allowed:
    FOR_ONE_BITS(b,n,p,w)
      user statement;
  
  The user should not change pp or ww.
  The user can do BIT_RESET(bb,pp) during the loop if he wants.
*/
#define FOR_ONE_BITS(bb,nn,pp,ww) \
  for ((ww)=0, (pp)=0; (pp)<(nn); ((ww) += (((++(pp))&BIT_RMASK_POW2)==0))) { \
    BITWORD ourw;                                                             \
    const int ourwordlen = BIT_WORDS(nn);                                     \
    for ( ourw = (bb)[(ww)] & ~ bitword_rmask[ (pp) & BIT_RMASK_POW2 ];       \
          ((ww) < ourwordlen) && (IS_BITWORD_ALL_ZERO(ourw));                 \
          ourw = (bb)[++(ww)] ) {}                                            \
    if ((ww) >= ourwordlen)  break;                                           \
    (pp) = (ww)*BITS_PER_WORD + bitword_find_smallest(ourw); 
    /* ( user code goes here; pp gives position of current bit ) */

#ifdef OLD
#define FOR_ONE_BITS(bb,nn,pp) \
    for ((pp)=0; ((pp)=bit_find_smallest((bb),(pp),(nn)))!=-1; (pp)++)
#endif /*OLD*/


void bit_global_init P(( ));
BITWORD **mat_bit_alloc P(( const int, const int, void *(*) P((const int)) ));
void mat_bit_free P(( BITWORD **, void (*) P((void *)) ));
void bit_print_hex P(( FILE *, BITWORD *, const int ));
void bit_print_binary P(( FILE *, const BITWORD *, const int ));
void bit_make_vec P(( const BITWORD *, const int, int *, int * ));
void bit_make_from_matl
  P(( BITWORD *, int **, const int, const int *, const int ));
int bit_find_by_rank P(( const BITWORD *, const int, const int ));
int bit_compare P(( const BITWORD *, const BITWORD *, const int ));


/**************************************************************************
  inline functions (be careful about changing the order of func. definition)
**************************************************************************/

/* declarations (included for compatibility with non-gnuc compilers) */

#ifndef __GNUC__
BITWORD * bit_alloc P(( const int, void *(*) P((const int)) ));
void bit_free P(( BITWORD *, void (*) P((void *)) ));
void bit_zero P(( BITWORD *, const int ));
int bit_is_zero P(( const BITWORD *, const int ));
void bit_ones P(( BITWORD *, const int ));
void bit_and P(( const BITWORD *, const BITWORD *, BITWORD *, words ));
int  bit_count P(( const BITWORD *, const int ));
void bit_copy P(( BITWORD *, const BITWORD *, const int ));
void bit_and P(( BITWORD *, const int, const BITWORD *, const BITWORD * ));
void bit_or P(( BITWORD *, const int, const BITWORD *, const BITWORD * ));
void bit_xor P(( BITWORD *, const int, const BITWORD *, const BITWORD * ));
void bit_not P(( BITWORD *, const int, const BITWORD * ));
void bit_minus P(( BITWORD *, const int, const BITWORD *, const BITWORD * ));
void bit_make_from_vec P(( BITWORD *, const int, const int *, const int ));
void bit_make_from_subset P(( const BITWORD *, const int, const int *,
                              const int, BITWORD *, BITWORD * ));
int bit_find_smallest P(( const BITWORD *, const int, const int ));
int bit_find_largest P(( const BITWORD *, const int ));
void bit_permute P(( BITWORD *const, const BITWORD *const,
                      const int, const int * const ));
#endif  /*__GNUC__*/

/*  note: inline is defined to the empty string unless we're using gnuc */
#undef  INLINE
#ifdef  SRCFILE_LIBPROG_BITARRAY
#define INLINE  __inline__
#else
#define INLINE  __inline__ extern
#endif

#if defined(__GNUC__) || defined(SRCFILE_LIBPROG_BITARRAY)

/*  Allocate a bit array that can hold n bits.
    //////////////////////////////////////////////////////////////////////
    //////////////////////////////////////////////////////////////////////
    /// NOTE: Recall that the unused bits of a bit array should be 0.  Most
    /// callers of bit_alloc() should call bit_zero() or bit_ones() immediately
    /// after calling this function, or set the unused bits of the newly
    /// allocated bit array to 0.  (For the latter, see bit_ones() and the
    /// use of the BIT_LASTMASK macro.)
    //////////////////////////////////////////////////////////////////////
    //////////////////////////////////////////////////////////////////////
*/
INLINE BITWORD *
bit_alloc(
const int n,
void  *(*allocfunc) P((const int))
)
{
  return( (*allocfunc) (BIT_WORDS(n) * sizeof(BITWORD)) );
}

INLINE void
bit_free(
BITWORD   *a,
void      (*freefunc) P((void *))
)
{
  (*freefunc) (a);
}

/*  Zero a bit array of n bits
    It is not necessary to call bit_global_init() before calling this function.
*/
INLINE void
bit_zero(
BITWORD *a,
const int n
)
{
  const int words = BIT_WORDS(n);
  int i;
  /*CCOULDO: memzero, maybe */
  for (i = 0; i < words; i++)  a[i] = 0;
}

/*  Determine if a bit array is all-bits zero.  Returns:
      0   some bit is set
      1   all bits are zero
*/
INLINE int
bit_is_zero(
const BITWORD *a,
const int n
)
{
  const int words = BIT_WORDS(n);
  int i;

  for (i = 0; i < words; i++) {
    if (IS_BITWORD_NOT_ALL_ZERO(a[i]))  return(0);
  }
  return(1);
}

/*  Set a bit array of n bits to all ones.  Bits that are part of the
    word array but not part of the bit array are set to 0.
*/
INLINE void
bit_ones(
BITWORD *a,
const int n
)
{
  if (n > 0) {
    const int words = BIT_WORDS(n);
    int i;
    /* COULDDO: memset, maybe */
    for (i = 0; i < words-1; i++)  a[i] = 0xffffffff;
    a[words-1] = BIT_LASTMASK(n);
  }
}

/*  Count the number of bits set in an n-bit array.
    Counts bits that are part of the word array but not part of the
    bit array (so these bits should be 0.)
    Assumes the total count is small enough to fit in an int.
*/
INLINE int
bit_count(
const BITWORD *a,
const int n
)
{
  const int words = BIT_WORDS(n);
  int i,count;

  for (count = 0, i = 0; i < words; i++) {
    count += bitword_count(a[i]);
  }
  return(count);
}

/*  Do the assignment "dest" <-- "src".  dest and src are bit vectors
    that use the same number of words.
*/
INLINE void
bit_copy(
BITWORD       *dest,
const BITWORD *src,
const int     n
)
{
  const int words = BIT_WORDS(n);
  int i;

  for (i = 0; i < words; i++)  dest[i] = src[i];
}


/*  Determine the AND of two n-bit strings (set c = a AND b )
    It is allowable to have c == a or c == b.
*/
INLINE void
bit_and(
BITWORD *c,
const int n,
const BITWORD *a,
const BITWORD *b
)
{
  const int words = BIT_WORDS(n);
  int i;

  for (i = 0; i < words; i++)  c[i] = a[i] & b[i];
}

/*  Determine the OR of two n-bit strings (set c = a OR b )
    It is allowable to have c == a or c == b.
*/
INLINE void
bit_or(
BITWORD *c,
const int n,
const BITWORD *a,
const BITWORD *b
)
{
  const int words = BIT_WORDS(n);
  int i;

  for (i = 0; i < words; i++)  c[i] = a[i] | b[i];
}

/*  Determine the XOR of two n-bit strings (set c = a XOR b )
    It is allowable to have c == a or c == b.
*/
INLINE void
bit_xor(
BITWORD *c,
const int n,
const BITWORD *a,
const BITWORD *b
)
{
  int  i;
  const int words = BIT_WORDS(n);

  for (i = 0; i < words; i++)  c[i] = a[i] ^ b[i];
}

/*  Determine the NOT of an n-bit string (set b = NOT a)
    It is allowable to have b == a.
*/
INLINE void
bit_not(
BITWORD *b,
const int n,
const BITWORD *a
)
{
  const int words = BIT_WORDS(n);
  int  i;

  for (i = 0; i < words; i++)  b[i] = ~ a[i];
  b[words-1] &= BIT_LASTMASK(n);
}

/*  Determine a-b (set c to 1 where a=1 and b=0)
*/
INLINE void
bit_minus(
BITWORD *c,
const int n,
const BITWORD *a,
const BITWORD *b
)
{
  int  i;
  const int words = BIT_WORDS(n);

  /* note that the last bits of a are already 0 */
  for (i = 0; i < words; i++)  c[i] = a[i] & (~ b[i]);
}

/*  Make a bit array that has a 1 in position i if element i appears in
    list[0..listlen-1].

    n must be big enough to handle the largest element X that appears
    in the list, i.e., we need n >= X+1.

    Is compatible with bit_make_vec().
*/
INLINE void
bit_make_from_vec(
BITWORD *a,
const int n,
const int *list,
const int listlen
)
{
  int i;

  bit_zero(a, n);
  for (i = 0; i < listlen; i++)
    BIT_SET(a, list[i]);
}

/*  Given a bit array that describes a universe U, and a vector V that
    describes a subset of U, construct bit arrays for V and U-V.
*/
INLINE void
bit_make_from_subset(
const BITWORD *U,       /* universe from which V is drawn */
const int n,            /* length of all bit arrays involved */
const int *V,           /* list describing a subset of U */
const int V_len,        /* length of V[] */
BITWORD *V_mask_ret,    /* bit array for V; caller-allocated length is n */
BITWORD *non_V_mask_ret /* bit array for U-V; caller-allocated length is n */
)
{
  bit_make_from_vec(V_mask_ret, n, V, V_len);
  bit_not(non_V_mask_ret, n, V_mask_ret);
  bit_and(non_V_mask_ret, n, non_V_mask_ret, U);
}

/*  Find the smallest bit position that is nonzero.  This is similar to
    bit_find_smallest().

    Returns:
      -1      if bit array is 0 in all positions at start_pos and larger
      p >= 0  smallest index p such that bit in position p is nonzero,
              such that p >= start_pos.

    The array a is not changed by this function.
    
    See also the FOR_ONE_BITS macro.
*/
INLINE int
bit_find_smallest(
const BITWORD *a,
const int start_pos,    /* smallest bit to look at; in range 0 .. len-1 */
const int len           /* bit length of a */
)
{
  if (start_pos >= len)  return(-1);    /* also catches len == 0 */
  {
    const int words = BIT_WORDS(len);
    int word;
    BITWORD w;

    word = BIT_WORD_INDEX(start_pos);
    w = a[word] & ~ bitword_rmask[ start_pos & BIT_RMASK_POW2 ];

    while (1) {
      if (IS_BITWORD_NOT_ALL_ZERO(w))
        return( word*BITS_PER_WORD + bitword_find_smallest(w) );
      word++;
      if (word == words)  break;
      w = a[word];
    }
    return(-1);
  }
}

/*  Find the largest bit position that is nonzero.  This is similar to
    bit_find_smallest().

    Returns:
      -1      if bit array is 0 in all positions at start_pos and smaller
      p >= 0  largest index p such that bit in position p is nonzero,
              such that p <= start_pos.

    The array a is not changed by this function.
*/
INLINE int
bit_find_largest(
const BITWORD *a,
const int start_pos /* first bit to look at; in range 0 .. len-1 */
                    /* (the array is assumed to have length len>= start_pos) */
)
{
  int word;
  BITWORD w;

  ASSERT(start_pos >= 0);

  word = BIT_WORD_INDEX(start_pos);
  w = a[word] & ( (bitword_rmask+1) [ start_pos & BIT_RMASK_POW2 ] );
#ifdef NODEBUG
  printf("init: start_pos=%d  mask=%08x\n",
    start_pos, (bitword_rmask+1)[start_pos&BIT_RMASK_POW2] );
#endif

  while (1) {
#ifdef NODEBUG
    printf("loop: word=%2d  w=%08d\n", word, w);
#endif
    if (IS_BITWORD_NOT_ALL_ZERO(w))
      return( word*BITS_PER_WORD + bitword_find_largest(w) );
    if (word == 0)  break;
    word--;
    w = a[word];
  }
  return(-1);
}

/*COULDDO: functions like find_smallest and find_largest that take an
  arbitrary of the bit array; the existing functions allow ranges to the
  left or right side of an array, but not in the middle, i.e. you cannot
  look for the leftmost bit in positions 2..7 if the array has length 10.
*/

/*  Apply a given permutation to the bits of a bit array.
    Bit i in the original array is mapped onto bit perm[i] in the new array.

    It is not necessary to call bit_global_init() before calling this function.
*/
INLINE void
bit_permute(
BITWORD *const dest,
const BITWORD *const src,
const int n,
const int * const perm        /* bit i is mapped to bit perm[i] */
)
{
  int i;

  bit_zero(dest, n);
  for (i = 0; i < n; i++) {
    if (IS_BIT_SET(src,i))
      BIT_SET(dest,perm[i]);
  }
}


#endif  /* defined(__GNUC__) || defined(source file) */

#endif /* _301093_180645 */
