/*
  -------------------- file bitword.h
  Created 12 May 92 by Andrew Shapira, Rensselaer Polytechnic Institute.
  Bit manipulation macros and inline functions that operate on a word
  (a BITWORD).

	Copyright (C) 1992 Andrew Shapira.

	This file is part of the Ntuplib library.

	Ntuplib is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	Ntuplib is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with the Ntuplib Library; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


/*
  Assumes BITWORD == unsigned int and sizeof(BITWORD) == 4.
  If you change this, you may have to change other functions.  If you move
  to a machine where sizeof(unsigned int != 32), probably the easiest
  thing to do is to change "BITWORD" into a type that is 32 bits long.

  USERS MUST CALL bitword_global_init() before calling these routines.
*/


#ifndef _020992_123651_bitword
#define _020992_123651_bitword

#include "common.h"

#define BITWORD               unsigned int

extern BITWORD bitword_rmask[];


/* set and reset bit "bb" of bitword "ww".  The bits are numbered
    from right to left starting at 0.  The caller should avoid depending
    on how the bits are ordered.
    (See bitword_find_smallest and bitword_find_largest.)
*/
#define BITWORD_SET(ww,bb)        ((ww) |= ((BITWORD) 1 << (bb)))
#define BITWORD_RESET(ww,bb)      ((ww) &= ~((BITWORD) 1 << (bb)))
/* return 0 if bb is not set, else return nonzero */
#define BITWORD_IS_BIT_SET(ww,bb) ((ww) & ((BITWORD) 1 << (bb)) )

#define ALL_BITS_ONE                (0xffffffff)
#define ALL_BITS_ZERO               (0x00000000)
/*  Return 1 if bitword is all-bits-0, return 0 otherwise.
    This is the preferred method for determining whether a word is 0.
    Note: some functions in this library might not use this method
    (it would be better if they did, though).
    MACHINE DEPENDENT.
*/
#define IS_BITWORD_ALL_ZERO(ww)         ((ww) == ALL_BITS_ZERO)
#define IS_BITWORD_ALL_ONE(ww)          ((ww) == ALL_BITS_ONE)
/* Return 0 if all-bits-0, else nonzero.  MACHINE DEPENDNT
*/
#define IS_BITWORD_NOT_ALL_ZERO(ww)     ((ww) != ALL_BITS_ZERO)


void bitword_global_init P(( ));
void bitword_make_rmask P(( BITWORD *, const int ));

void bitword_count_table_init P((void));
void bitword_count_table_free P((void));
void bitword_small_table_init P((void));
void bitword_small_table_free P((void));
void bitword_large_table_init P((void));
void bitword_large_table_free P((void));
void bitword_rev_table_alloc P(( const int, void  *(*) P((const int)) ));
void bitword_rev_table_init P((void));
void bitword_rev_table_free P(( void  (*) P((void *)) ));

/**************************************************************************
  inline functions (be careful about changing the order of func. definition)
**************************************************************************/

/* declarations (included for compatibility with non-gnuc compilers) */

#ifndef __GNUC__
int bitword_find_smallest P(( const BITWORD ));
int bitword_find_largest P(( const BITWORD ));
int bitword_count P(( const BITWORD ));
int bitword_rev P(( const BITWORD ));
int bitword_compare P(( const BITWORD, const BITWORD ));
#endif  /*__GNUC__*/

/*  note: inline is defined to the empty string unless we're using gnuc */
#undef  INLINE
#ifdef  SRCFILE_LIBPROG_BITWORD
#define INLINE  __inline__
#else
#define INLINE  __inline__ extern
#endif

#if defined(__GNUC__) || defined(SRCFILE_LIBPROG_BITWORD)

/* 
  Return the number of bits set in a 32-bit word.
*/
INLINE int
bitword_count(
const BITWORD w
)
{
  extern unsigned char *_bitword_count_table;
                                  /* [i]: num. of bits set, i = 0, 2^16-1 */

  return( _bitword_count_table[(w >> 16) & 0x0000ffff ] +
          _bitword_count_table[        w & 0x0000ffff ]);
}

/* 
  Return the position of the smallest bit.  Returns -1 if no bits are set.
*/
INLINE int
bitword_find_smallest(
const BITWORD w
)
{
  extern char *_bitword_small_table;  /* smallest bit of 16-bit patterns */

  if (w & 0x0000ffff) return(_bitword_small_table[w & 0x0000ffff]);
  if (w)              return(16 + _bitword_small_table[(w >> 16) & 0x0000ffff]);
  return(-1);
}

/* 
  Return the position of the largest bit.  Returns -1 if no bits are set.
*/
INLINE int
bitword_find_largest(
const BITWORD w
)
{
  extern char *_bitword_large_table;  /* largest bit of 16-bit patterns */

  if (w & 0xffff0000) return(16 + _bitword_large_table[(w >> 16) & 0x0000ffff]);
  if (w)              return(_bitword_large_table[w]);
  return(-1);
}


/*  Reverse the smallest n bits of a given 32 bit word.  n is specified
    by calling _bitword_rev_table_alloc().
    You must have previously called _bitword_rev_table_alloc() and
    _bitword_rev_table_init().
*/
INLINE int
bitword_rev(
const BITWORD w
)
{
  extern BITWORD *_bitword_rev_table;
  extern int _bitword_rev_len;

  if (_bitword_rev_len <= 16) {
    ASSERT((w & 0xffff0000) == 0);
    return(_bitword_rev_table[w]);
  }
  else {
    extern int _bitword_rev_len_m_16;
    extern int _bitword_32_m_rev_len;

    return( (_bitword_rev_table[w & 0x0000ffff] << _bitword_rev_len_m_16) |
            (_bitword_rev_table[(w>>16) & 0x0000ffff] >> _bitword_32_m_rev_len) );
  }
}

/*
  Compares two BITWORDS.
  Returns:
      -1      x < y
      0       x = y
      1       x > y
    Is compatible with bitword_find_smallest() and bitword_find_largest().
    Can be used as comparison function to qsort().
*/
INLINE int
bitword_compare(
const BITWORD x,
const BITWORD y
)
{
  if (x == y)  return(0);
  {
    const BITWORD xor = x ^ y;
    const pos = bitword_find_smallest(xor); /* first pos that x and y differ */
    ASSERT(pos != -1);
    if (BITWORD_IS_BIT_SET(x,pos))  return(1);
    return(-1);
  }
}

#endif  /* defined(__GNUC__) || defined(source file) */

#endif /* _020992_123651_bitword */
