/*
  -------------------- file matrix.c
  Created Spring 92 by Andrew Shapira, Rensselaer Polytechnic Institute.
  Routines for rectangular matrix manipulation (see also matvec.c).

	Copyright (C) 1992 Andrew Shapira.

	This file is part of the Ntuplib library.

	Ntuplib is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	Ntuplib is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with the Ntuplib Library; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define SRCFILE_LIBPROG_MATRIX

#include <stdio.h>
#include "our_common.h"
#include "getnum.h"
#include "io.h"
#include "matrix.h"
#include "memalloc.h"
#include "permute.h"
#include "random.h"
#include "util.h"
#include "vector.h"


/*  There are often two versions of a routine, one for matrices of
    bytes (unsigned char) and one for matrices of arrays of ints.  The
    convention is to place the two routines next to each other in the
    source, with the integer routine first.
*/


/*  set result[i][j] to be the number of ocurences of "j" in row "i" of m.
    Assumes that the values of m[][] are in the range 0 .. n-1.
*/
void
mat_row_histogram(
int         * const * const m,
const int   n,
int         * const * const result
)
{
  int   mrow,mcol;

  mat_zero(result,n);
  for (mrow = 0; mrow < n; mrow++) {
    for (mcol = 0; mcol < n; mcol++) {
      ASSERT((m[mrow][mcol] >= 0) && (m[mrow][mcol] <= n-1));
      result[mrow][m[mrow][mcol]]++;
    }
  }
}

/*  set result[i][j] to be the number of ocurences of "j" in column "i" of m.
    Assumes that the values of m[][] are in the range 0 .. n-1.
*/
void
mat_col_histogram(
int         * const * const m,
const int   n,
int         * const * const result
)
{
  int   mrow,mcol;

  mat_zero(result,n);
  for (mcol = 0; mcol < n; mcol++) {
    for (mrow = 0; mrow < n; mrow++) {
      ASSERT((m[mrow][mcol] >= 0) && (m[mrow][mcol] <= n-1));
      result[mcol][m[mrow][mcol]]++;
    }
  }
}

/*  Set h[i] to the number of times "min+i" occurs in "m".  (Find the
    distribution of the elements of m.)

    The reason for having the caller supply min and max is that it encourages
    the caller to have the correct length of "h".  We could easily find
    min and max in this function, but this is better.

    h[i] is set to the number of occurrences of value "min+i" in "m", for all
    i, 0 <= i <= max-min
*/
void
mat_uchar_histogram(
unsigned char   **m,    /* input matrix */
const int       rows,   /* size of m */
const int       cols,
const int       min,    /* minimum and maximum values actually occurring in m */
const int       max,
int             *h      /* histogram to be returned */
                        /* must have length max-min+1 or more */
)
{
  const int len = max-min+1;
  int   row,col,v;

#ifdef  HDEBUG
  printf("mat_uchar_histogram: finding histogram of: \n");
  printf("rows=%d  cols=%d  min=%d  max=%d  len=%d\n",
    rows, cols, min, max, len);
  mat_uchar_print2(m, rows, cols);
#endif  /*HDEBUG*/
  vec_zero(h, len);
  for (row = 0; row < rows; row++) {
    for (col = 0; col < cols; col++) {
      v = m[row][col];
      ASSERT((v >= min) && (v <= max));
      h[v-min]++;
    }
  }
}

void
mat_print2(
FILE      *f,
int       * const * const m,
const int rows,
const int cols
)
{
  int   row, col, min, max, width, outcol;

  /* determine field length */
  mat_minmax2(m,rows,cols,&min,&max);
  width = MAX(fieldlen(min),fieldlen(max))+1;

  for (row = 0; row < rows; row++) {
    outcol = 0;
    for (col = 0; col < cols; col++) {
      if (outcol+width > OUTPUT_COLUMNS) {
        fprintf(f, "\n");
        spaces_print(f, width);
        outcol = width;
      }
      fprintf(f, "%*d", width, m[row][col]);
      outcol += width;
    }
    fprintf(f, "\n");
  }
  if (rows == 0)
    fprintf(f, "(empty)\n");
}

void
mat_uchar_print2(
FILE          *f,
unsigned char * const * const m,
const int     rows,
const int     cols
)
{
  int   row, col;
  int   fill;
  int   min,max;
  int   width;
  char *buff;

  bsbegin() ;
  buff = bsalloc(cols*2+13);

  mat_uchar_minmax2(m,rows,cols,&min,&max);
  if ((min >= 0) && (max <= 9)) {
    for (fill = 0; fill < cols-1; fill++)
      buff[2*fill+1] = ' ';
    if (cols > 0) {
      buff[2*cols-1] = '\n';
      buff[2*cols] = '\0';
    }
    else {    /* n by 0 matrix; might save us some time */
      buff[0] = '\n';
      buff[1] = '\0';
    }
    for (row = 0; row < rows; row++) {
      for (col = 0; col < cols; col++)
        buff[2*col] = m[row][col] + '0';
      fputs(buff, f);
    }
  }
  else {
    width = MAX(fieldlen(min), fieldlen(max))+1;
  
    for (row = 0; row < rows; row++) {
      for (col = 0; col < cols; col++)
        fprintf(f, "%*d", width, m[row][col]);
      fprintf(f, "\n");
    }
  }
  if ((rows == 0) || (cols == 0))
    printf("(empty)\n");
  bsend();
}

void
mat_double_print2(
FILE      *f,
double    * const * const m,
const int rows,
const int cols,
const int rightofdecimal  /* how many places to show to right of decimal pt. */
)
{
  int       row, col, outcol;
  double    min,max;
  int       leftofdecimal, width;

  /* determine field length */
  mat_double_minmax2(m,rows,cols,&min,&max);
  leftofdecimal = MAX(fieldlen((int) min),fieldlen((int) max));
  width = leftofdecimal+rightofdecimal+2;

  for (row = 0; row < rows; row++) {
    outcol = 0;
    for (col = 0; col < cols; col++) {
      if (outcol+width > OUTPUT_COLUMNS) {
        fprintf(f, "\n");
        spaces_print(f, width);
        outcol = width;
      }
      fprintf(f, "%*.*f", width, rightofdecimal, m[row][col]);
      outcol += width;
    }
    fprintf(f, "\n");
  }
  if (rows == 0)
    fprintf(f, "(empty)\n");
}

void
mat_print(
FILE      *f,
int       * const * const m,
const int n
)
{
  mat_print2(f,m,n,n);
}

void
mat_uchar_print(
FILE          *f,
unsigned char * const * const m,
const int     n
)
{
  mat_uchar_print2(f,m,n,n);
}

/* ------------ Begin support for mat_print_listformat */

struct mat_listformat_struct {
  int   width;
  int   default_row_length;
  int   *row_lengths;
  int   *const *m;
};

static int
listformat_component_compare(
const void *p1,             /*  const int ** */
const void *p2,             /*  const int ** */
const void *printinfoin     /*  struct mat_listformat_struct *  */
)
{
  int         * const * v1 = p1;
  int         * const * v2 = p2;
  int         index1, index2;
  int         len1, len2;
  const struct mat_listformat_struct     *printinfo = printinfoin;
  const int   *row_lengths = printinfo->row_lengths;
  int         n;

  if (row_lengths) {
    index1 = v1 - printinfo->m;
    len1 = row_lengths[index1];
    index2 = v2 - printinfo->m;
    len2 = row_lengths[index2];
    if (len2 != len1)
      return(len1 - len2);
    n = len1;
  }
  else
    n = printinfo->default_row_length;

  return(vec_cmp(*v1, *v2, n));
}

static int
mat_listformat_component_print(
const void  *element,       /*  const int **  */
const void  *printinfoin,   /*  struct mat_listformat_struct *  */
const int   col
)
{
  int         * const * vec = element;
  const struct mat_listformat_struct    *printinfo = printinfoin;
  const int   width = printinfo->width;
  const int   default_row_length = printinfo->default_row_length;
  int         rowlen;
  const int   *row_lengths = printinfo->row_lengths;
  int         row;
  int         vec_printinfo[1];

  if (row_lengths) {
    row = vec - printinfo->m;
    rowlen = row_lengths[row];
  }
  else
    rowlen = default_row_length;

  vec_printinfo[0] = width;

  return(list_print(
    stdout, *vec, sizeof(int), rowlen, 1, 1, col, 1, OUTPUT_COLUMNS, width,
    & vec_printinfo[0], list_int_print, list_int_compare));
}

/* ------------ End support routines for mat_print_listformat. */


/*  Print a matrix in list format.
    If row_lengths == NULL: all rows are of length default_row_length
    If row_lengths != NULL: row i is of length row_length[i]
*/
int
mat_print_listformat(
FILE      *f,
int       * const * const m,
const int rows,               /*  number of rows in matrix  */
const int default_row_length, /*  what to use if row_lengths == NULL */
int       *row_lengths        /*  if not null, [i] is entries in row i */
)
{
  const int                       leftmargin = 1;
  /* print information for mat_listformat_component_print() and compare */
  struct mat_listformat_struct    printinfo;
  int                             min,max;
  int                             width;
  int                             row,col,val,init,cols;

  /*  Calculate field length.  */
  init = 1;
  min=max=0;
  for (row = 0; row < rows; row++) {
    if (row_lengths)
      cols = row_lengths[row];
    else
      cols = default_row_length;
    for (col = 0; col < cols; col++) {
      val = m[row][col];
      if (! init) {
        min = MIN(val, min);
        max = MAX(val, max);
      }
      else {
        min = max = val;
        init = 0;
      }
    }
  }
  /*  Fill in printinfo fields.  */
  width = MAX(fieldlen(min),fieldlen(max));
  printinfo.width = width;
  printinfo.default_row_length = default_row_length;
  printinfo.row_lengths = row_lengths;
  printinfo.m = m;

  return(  list_print( f, m, sizeof(int *), rows, 0, 0, 1, leftmargin,
              OUTPUT_COLUMNS, width, &printinfo,
              mat_listformat_component_print, listformat_component_compare)  );
}

/* Returns 1 if successful, 0 if not */
int
mat_uchar_read2(
FILE *f,
unsigned char **m,          /* m is m[d1size][d2size] */
const int d1size,
const int d2size
)
{
  int           d1,d2,i;

  for (d1 = 0; d1 < d1size; d1++) {
    for (d2 = 0; d2 < d2size; d2++) {
      if (! getnum(f, & i)) return(0);
      if ((i < 0) || (i > 255)) return(0);
      m[d1][d2] = i;
    }
  }
  return(1);
}

/* like mat_uchar_read2, but we read number of rows and columns first.
  If we allocate space, then *m_ret is set to be non-NULL.
  Returns 1 if successful, 0 if failed.
  If we failed, then:
    If m_ret is not NULL, then it contains whatever we read before the error.
*/
int
mat_uchar_read2_with_dim(
FILE *f,
unsigned char ***m_ret,
int *rows_ret,
int *cols_ret,
void  *(*allocfunc) P((const int))
)
{
  *m_ret = NULL;
  if (! getnum(f,rows_ret)) return(0);
  if (! getnum(f,cols_ret)) return(0);
  if ((*rows_ret < 1) || (*cols_ret < 1))  return(0);
  *m_ret = mat_uchar_alloc2(*rows_ret,*cols_ret,allocfunc);
  return(mat_uchar_read2(f,*m_ret,*rows_ret,*cols_ret));
}

int
mat_read2(
FILE *f,
int **m,          /* m is m[d1size][d2size] */
const int d1size,
const int d2size
)
{
  int           d1,d2,i;

  for (d1 = 0; d1 < d1size; d1++) {
    for (d2 = 0; d2 < d2size; d2++) {
      if (! getnum(f, & i))
        return(0);
      m[d1][d2] = i;
    }
  }
  return(1);
}

int
mat_uchar_read(
FILE *f,
unsigned char **m,
const int n
)
{
  return(mat_uchar_read2(f, m,n,n));
}

int
mat_read(
FILE *f,
int **m,
const int n
)
{
  return(mat_read2(f, m,n,n));
}

int
mat_uchar_is_main_diag0(
unsigned char       **m,
const int           n
)
{
  int   row;

  for (row = 0; row < n; row++) {
    if (m[row][row])
      return(0);
  }
  return(1);
}

int
mat_is_main_diag0(
int         **m,
const int   n
)
{
  int   row;

  for (row = 0; row < n; row++) {
    if (m[row][row])
      return(0);
  }
  return(1);
}

int
mat_uchar_is_main_diag1(
unsigned char   **m,
const int       n
)
{
  int   row;

  for (row = 0; row < n; row++) {
    if (m[row][row] != 1)
      return(0);
  }
  return(1);
}

int
mat_is_main_diag1(
int       **m,
const int n
)
{
  int   row;

  for (row = 0; row < n; row++) {
    if (m[row][row] != 1)
      return(0);
  }
  return(1);
}

/*  Matrix version of vec_table_to_vec.
*/
void
mat_table_to_mat(
int         **table,        /*  const int **  */
const int   rows,           /*  number of rows (matrix might not be square) */
const int   cols,
const int   min,
const int   end_of_row_mark,
int         **mret          /*  should be rows by (max-min+2)  */
)
{
  int   row;

  for (row = 0; row < rows; row++)
    vec_table_to_vec(table[row],min,cols,end_of_row_mark,mret[row]);
}

/*  Scale the values in a matrix to be in the range desired_min .. desired_max.
  DEBUG: it would be better to convert the table beforehand, and then use
  apply_map().
  DEBUG: try to get rid of MIN and MAX (kludgy)
*/
void
mat_scale2(
int   **m,
const int   rows,
const int   cols,
const int   desired_min,
const int   desired_max
)
{
  int   row,col;
  int   min,max;
  int   v;

  ASSERT(desired_min <= desired_max);
  mat_minmax2(m,rows,cols,&min,&max);
  
  if (max != min) {
    for (row = 0; row < rows; row++) {
      for (col = 0; col < cols; col++) {
        v = ((m[row][col] - min) * (desired_max - desired_min) / (max - min)) +
          desired_min;
        v = MAX(v, desired_min);
        m[row][col] = MIN(v, desired_max);
      }
    }
  }
  else
    mat_set2(m,rows,cols,(desired_min + desired_max) / 2);
}

void
mat_uchar_scale2(
unsigned char   **m,
const int   rows,
const int   cols,
const int   desired_min,
const int   desired_max
)
{
  int   row,col;
  int   min,max;
  int   v;

  ASSERT(desired_min <= desired_max);
  ASSERT((desired_min >= 0) && (desired_min <= 255));

  mat_uchar_minmax2(m,rows,cols,&min,&max);
  
  if (max != min) {
    for (row = 0; row < rows; row++) {
      for (col = 0; col < cols; col++) {
        v = ((m[row][col] - min) * (desired_max - desired_min) / (max - min)) +
          desired_min;
        v = MAX(v, desired_min);
        m[row][col] = MIN(v, desired_max);
      }
    }
  }
  else
    mat_uchar_set2(m,rows,cols,(desired_min + desired_max) / 2);
}

void
mat_double_scale2(
double          **m,
const int       rows,
const int       cols,
const double    desired_min,
const double    desired_max
)
{
  int   row,col;
  double    min,max;
  double    v;

  ASSERT(desired_min <= desired_max);
  mat_double_minmax2(m,rows,cols,&min,&max);
  
  if (max != min) {
    for (row = 0; row < rows; row++) {
      for (col = 0; col < cols; col++) {
        v = ((m[row][col] - min) * (desired_max - desired_min) / (max - min)) +
          desired_min;
        v = MAX(v, desired_min);
        m[row][col] = MIN(v, desired_max);
      }
    }
  }
  else
    mat_double_set2(m,rows,cols,(desired_min + desired_max) / 2);
}


/*  Generate a given permutation of a matrix.  If we view the matrix as
    the adjacency matrix of a graph, then what this function does is
    equivalent to applying a given permutation of the vertices.
*/
void
mat_uchar_permute(
unsigned char **dest,
unsigned char **source, /* const */
const int n,
int *permute_vector     /* length n vector, maps [i] to permute_vector[i] */
)
{
  int   i,j;

  for (i = 0; i < n; i++) {
    const unsigned char *sourcerow = source[i];
    char *destrow;
    destrow = dest[permute_vector[i]];
    for (j = 0; j < n; j++) {
      destrow[permute_vector[j]] = sourcerow[j];
    }
  }
}

/*  Generate a random permutation of a matrix.  If we view the matrix as
    the adjacency matrix of a graph, then what this function does is
    equivalent to randomly permuting the vertices.
*/
void
mat_uchar_permute_random(
unsigned char **m,
const int n
)
{
  unsigned char **copied;
  int *permute_vector;
  long random();

  bsbegin();

  copied = mat_uchar_alloc(n, bsalloc);
  permute_vector = vec_alloc(n, bsalloc);

  randperm(permute_vector, n, random);
  mat_uchar_copy(copied, m, n);
  mat_uchar_permute(m, copied, n, permute_vector);

  bsend();
}
 
/*  Generate a random binary matrix.
    Each element is a 1 with probability r/s, and a 0 with probability 1-r/s.
    Undefined behavior if s==0 or r/s>1
*/
void
mat_uchar_random_binary2(
unsigned char **a,
const int rows,
const int cols,
const unsigned long r,
const unsigned long s
)
{
  int row,col;

  for (row = 0; row < rows; row++) {
    unsigned char *thisrow;
    thisrow = a[row];
    for (col = 0; col < cols; col++)
      thisrow[col] = coinflip_rat(r,s);   /* coinflip_rat in random.h */
  }
}

void
mat_uchar_random_binary(
unsigned char **a,
const int n,
const unsigned long r,
const unsigned long s
)
{
  mat_uchar_random_binary2(a,n,n,r,s);
}

/*  Trim a matrix by deleting as many surrounding background rows
    and columns as possible, so that no foreground color pixels are
    changed.  If all pixels are the background color, then the trimmed
    matrix consists of a 1 by 1 array, whose pixel is the background
    color.

    The user specifies a background color; the definition of the
    foreground is anything that is not the background color.

    The new trimmed matrix is returned in *a_ret, *rows_ret, *cols_ret.
    The original matrix is unchanged.
*/
void
mat_uchar_trim2(
unsigned char ** a,
const int nrows,
const int ncols,
const int bgcolor,
unsigned char ***a_ret,
int *rows_ret,
int *cols_ret,
void  *(*allocfunc) P((const int))
)
{
  int toprow,botrow,leftcol,rightcol;
  int i,found;

  if ((nrows < 1) || (ncols < 1))  fatal("mat_uchar_trim2");

  /* find top row */
  for (found = 0, toprow = 0; toprow < nrows; toprow++) {
    for (i = 0; i < ncols; i++) {
      if (a[toprow][i] != bgcolor)
        {  found = 1;  break;  }
    }
    if (found)  break;
  }

  if (found) {    /* character is not all 0 */
    /* find bottom row */
    for (found = 0, botrow = nrows-1; botrow >= 0; botrow--) {
      for (i = 0; i < ncols; i++) {
        if (a[botrow][i] != bgcolor)
          {  found = 1;  break;  }
      }
      if (found)  break;
    }

    /* find left column */
    for (found = 0, leftcol = 0; leftcol < ncols; leftcol++) {
      for (i = 0; i < nrows; i++) {
        if (a[i][leftcol] != bgcolor)
          {  found = 1;  break;  }
      }
      if (found)  break;
    }

    /* find right column */
    for (found = 0, rightcol = ncols-1; rightcol >= 0; rightcol--) {
      for (i = 0; i < nrows; i++) {
        if (a[i][rightcol] != bgcolor)
          {  found = 1;  break;  }
      }
      if (found)  break;
    }
  }
  else {       /* character is all 0 */
    toprow = botrow = leftcol = rightcol = 0;   /* clone 0,0 pixel */
  }
  ASSERT((0 <= toprow) && (toprow <= botrow) && (botrow < nrows));
  ASSERT((0 <= leftcol) && (leftcol <= rightcol) && (rightcol < ncols));

  {
    const int newnrows = botrow - toprow + 1;
    const int newncols = rightcol - leftcol + 1;
    unsigned char **newa;
    int r,c;

    newa = mat_uchar_alloc2(newnrows,newncols,allocfunc);
    for (r = 0; r < newnrows; r++) {
      for (c = 0; c < newncols; c++) {
        newa[r][c] = a[toprow+r][leftcol+c];  /*COULDDO:speed*/
      }
    }
    *a_ret = newa;
    *rows_ret = newnrows;
    *cols_ret = newncols;
  }
}

/*  (Treat a matrix as an adjacency matrix.)
    Convert an adjacency matrix to a neighbor list.  Neighbor list is
    terminated by -1.  Nodes are not considered to be adjacent to
    themselves.  Works for directed and undirected graphs (has no
    dependence of the symmetry of the adjacency matrix).
*/
void
mat_uchar_neighbors(
unsigned char   * const * const m,
const int       n,                            /* m matrix is n*n */
int             * const * const neighbors_ret
)
{
  int   row,col;

  D(mat_zero(neighbors,n);) /* show up nicely for print statements */
  for (row = 0; row < n; row++) {
    unsigned char * const mrow = m[row];
    int * const neighbors_row = neighbors_ret[row];
    int howmany = 0;
    for (col = 0; col < n; col++) {
      ASSERT((row != col) || (mrow[col] == 0));
      if (mrow[col])  neighbors_row[howmany++] = col;
    }
    neighbors_row[howmany] = -1;
  }
}

int **
mat_uchar_alloc_neighbors(
unsigned char * const * const m,
const int n,
void  *(*allocfunc) P((const int))
)
{
  int **neighbors = mat_alloc(n, allocfunc);

  mat_uchar_neighbors(m,n,neighbors);
  return(neighbors);
}
