/*
  -------------------- file sort.c

	Copyright (C) 1992 Andrew Shapira.

	This file is part of the Ntuplib library.

	Ntuplib is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	Ntuplib is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with the Ntuplib Library; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#include <stdio.h>

/* COULDDO:
   qsort-like functions for heap sort.  Use a linear array to store
   the heap, with the tree implicitly defined by multiplying by 2.
*/

/*#define RI_DEBUG*/
/* #define DEBUG */
#include "our_common.h"

#include "alloca.h"
#include "matrix.h"
#include "memalloc.h"
#include "permute.h"
#include "sort.h"
#include "vector.h"

#ifdef  ISORT_COUNT
int isort_moves = 0;
#endif  /*ISORT_COUNT*/

void *memcpy P((void *, const void *, size_t));

/*  Throughout, it is assumed that the comparison function does not
    depend on the actual address of the objects being compared.
    The result of comparing two objects should be independent of
    where they are in memory.
*/




#define RANDOMIZE_SSORT

#define INTSORTBUFFLEN  1024
static int intsortbuff[INTSORTBUFFLEN];

/*  Sort a list that contains integers.  If you use this routine, be prepared
    to accept a statically allocated buffer of size (max-min+1).  If this
    is unacceptable, clone this routine and make a version that frees the
    buffer when it's done, or uses a global buffer but calls ssort() when
    the list is too big for the buffer.

    Takes O(maximum(max-min, n)) time if it uses the algorithm here;
    otherwise O(n log n) time if the algorithm decides to invoke ssort().

    Note: min and max do not have to be exact.

    Note: see the comment before bitarray.c/vec_make_from_bit().
*/
void
intsort(
int   *a,         /* array to be sorted */
int   n,          /* how many elements in array to sort (starting at a[0]) */
int   min,        /* no value in array is smaller than this */
int   max         /* no value in array is larger than this */
)
{
  register int  i,j,t;
  static int *locbuff;
  static int locbuffsize = 0;

D(printf("intsort: min=%d  max=%d\n   vec=", min, max);)
D(vec_print_int(a,n);)
  /*  If the list is small enough, it's better to use ssort()
      This formula is not known to be accurate.  It'd probably
      be better to use something like max-min+1 > K n lg n, where
      K is determined empirically for the current environment.

      TODO: could think about using bit sorting (vec_make_from_bit).
  */
  if (max-min+1 > 4*n)
    ssort((void *)(a),(n),sizeof(int),int_compare);

  /*  If we can, use the global variable intsortbuff
      for faster referencing and no malloc call.  If
      you change something here, change it below too.
  */
  else if (max-min+1 <= INTSORTBUFFLEN) {
    vec_zero(intsortbuff,max-min+1);
    for (i = 0; i < n; i++) {
      ASSERT((a[i] >= min) && (a[i] <= max));
      intsortbuff[a[i]-min]++;
    }
    for (i = 0, j = 0; j <= max-min; j++) {
      t = intsortbuff[j];
      while (t--) {
        ASSERT((i >= 0) && (i < n));
        a[i++] = j+min;
      }
    }
  }

  /*  Use local variable references.  This is an identical clone of the
      code above, except it first allocates memory (locbuff).  If you change
      something here, change it above too.
  */
  else {
    xalloc(locbuff,(max-min+1)*sizeof(int),locbuffsize);
    vec_zero(locbuff,max-min+1);
    for (i = 0; i < n; i++) {
      ASSERT((a[i] >= min) && (a[i] <= max));
      locbuff[a[i]-min]++;
    }
    for (i = 0, j = 0; j <= max-min; j++) {
      t = locbuff[j];
      while (t--) {
        ASSERT((i >= 0) && (i < n));
        a[i++] = j+min;
      }
    }
  }
}

/*  A function that can be used as the comparison routine for sorting
    integers using qsort(), or a function with the same arguments as
    qsort (such as ssort()).  Sorts into ascending order.
*/
int
int_compare(p1,p2)
const void  *p1,*p2;
{
  return(*(const int *)p1 - *(const int *)p2);
}

/*  Like int_compare, but sorts into descending order.
*/
int
int_compare_descending(p1,p2)
const void  *p1,*p2;
{
  return(*(const int *)p2 - *(const int *)p1);
}

/*  Binary search.
    Returns the index of "key".  If there are multiple copies of "key",
    returns an arbitrary one.  If "key" does not occur in the table,
    returns where "key" would occur.

    Cloned in pqueue lib / sarray.c.
*/
int
bfind(
const void    *key,                             /* element to possibly find */
void          *base,                            /* base of list */
const size_t  len,                              /* current length of list */
const size_t  width,                            /* size of each element */
int           (*compar) (const void *, const void *)  /* comparison function */
)
{
  char  *cbase = (char *) base;
  int   l,h,m;        /* lower, higher, middle end of current list range */
  int   compare;      /* result of comparison */

  l = 0;
  h = len-1;
  while (l <= h) {
    m = (l+h) / 2;

#ifdef  BDEBUG
    printf("l=%d  m=%d  h=%d  len=%d\n", l,m,h,(int) len);
#endif  /*BDEBUG*/
    compare = compar(key, (void *) (cbase+m*width));
    if (compare > 0)        l = m + 1;
    else if (compare < 0)   h = m - 1;
    else                    {  l = m;  break;  }
  }
  return(l);
}

/*  Insert an element into an already-sorted list.
    If the element is already in the list, it is not inserted, and we return
    -1.  Otherwise, we return the position where it was inserted.
    Differs from binsert_withdups() only in that duplicate elements are not
    inserted, and that the list length is updated automatically here.
    Any changes made here should be applied to binsert_withdups().

    Not tested recently.
*/
int
binsert_withoutdups(
const void    *key,                             /* element to possibly add */
void          *base,                            /* base of list */
size_t        *nelptr,                          /* location of list length */
const size_t  width,                            /* size of each element */
int           (*compar) (const void *, const void *)  /* comparison function */
)
{
  char  * const cbase = (char *) base;
  const int where = bfind(key,base,*nelptr,width,compar);
  if ( (*compar) (key, (void *) (cbase+where*width)) == 0 )
    return(-1);
  if (where < *nelptr)
    memmove(cbase+(where+1)*width,cbase+where*width, (*nelptr-where)*width);
  memcpy(cbase+where*width, key, width);
  (*nelptr)++;
  return(where);
}

/*  Insert an element into an already-sorted list.
    We return the position where it was inserted.
    Differs from binsert_withoutdups() only in that duplicate elements are
    inserted, and that the list length is not updated automatically.
    Any changes made here should be applied to binsert_withoutdups().
*/
int
binsert_withdups(
const void    *key,                             /* element to possibly add */
void          *base,                            /* base of list */
const size_t  len,                              /* current length of list */
const size_t  width,                            /* size of each element */
int           (*compar) (const void *, const void *)  /* comparison function */
)
{
  char  * const cbase = (char *) base;
  const int where = bfind(key,base,len,width,compar);
#ifdef  ISORT_COUNT
  isort_moves += len-where;
#endif  /*ISORT_COUNT*/
  if (where < len)
    memmove(cbase+(where+1)*width,cbase+where*width, (len-where)*width);
  memcpy(cbase+where*width, key, width);
  return(where);
}


/*  ---------- ssort()

    Shapira's randomized 3-partition quicksort / insertion sort hybrid.

    Same arguments as the library routine qsort().

    Possible ways to get improvements:
      - Optimize "SMALL"
      - Optimize the size of the group of elements used to determine the
          partition elements.
      - Optimize the number of partitions (currently fixed at 3).

    Written 29Nov91.
*/

/*  Returns boolean indicating whether element number xx is to be
    considered less than element number yy.
*/

/*  ----- macros for ssort()
*/

#define S_LESS(xx,yy)   (compar(cbase+(xx)*width,cbase+(yy)*width) < 0)

/*  Swap element number xx with element number yy.
*/
#define S_SWAP(xx,yy) do {  memcpy(swapbuff,cbase+(xx)*width,width); \
                            memcpy(cbase+(xx)*width,cbase+(yy)*width,width); \
                            memcpy(cbase+(yy)*width,swapbuff,width); \
                          } while (0)

#define S_PSWAP(xx,yy)  do { int t; t=newloc[(xx)]; \
                              newloc[(xx)]=newloc[(yy)]; newloc[(yy)]=t;  \
                              S_SWAP(xx,yy); \
                        } while (0);

/*  The size at which a collection of elements is sorted using insertion
    sort or direct element comparison.  To change SMALL, it is not enough
    to just change the constant; some code must change too.
*/
#define SMALL 16    /* must be at least 8 */


void
ssort(
void            *base,
const size_t    nel,
const size_t    width,
int   (*compar) P((const void *, const void *))
)
{
  char *cbase = base;               /* first element in entire list */
  
  int   *starts   = (int *) alloca(nel*sizeof(int));
  int   *ends     = (int *) alloca(nel*sizeof(int));
  char  *newloc   = (char *) alloca(nel);
  char  *swapbuff = (char *) alloca(width);
  int   onstack;                    /* how many entries currently on stack */
  int   first,last,size;            /* current area being sorted */
  int   howmany0,howmany1,howmany2; /* how many in each partition */
  /* rightmost element in each partition not known to be out of place */
  int   right0,right1,right2;
  int   fill0,fill1,fill2;          /* leftmost el. not known to be in place */
  int   i,j;
  int   needtosortfirst8;
  /* elements to compare vs. during partition-determining phase */
  int   compare1,compare2;
  long  random();                 /* used even w/o RANDOMIZE_SSORT */

#ifdef  SSORT_TEST_SUITE
  printf("ssort(.. , nel=%d , width=%d, ...)\n", nel, width);
#endif  /*SSORT_TEST_SUITE*/

#ifdef  RANDOMIZE_SSORT
  /*  Put the elements in a random order, so that the running time of the
      algorithm does not depend on the order of the input.
  */
  if (nel > SMALL) {
    /* AS: shift so as to reduce position bias from random() */
    const int   shift = random() % nel;

    for (i = 1; i < nel; i++) {
      j = random() % (i+1);
      if (j != i)
        S_SWAP((shift+i) % nel, (shift + j) % nel);
    }
  }
#endif  /*RANDOMIZE_SSORT*/

  /*  Initialize stack.  */
  onstack = 0;
  first = 0;
  last = nel-1;


  /**/
  needtosortfirst8 = 1;
  while (1) {
    size = last - first + 1;
#ifdef  SSORT_TEST_SUITE
    printf("ssort(), top of loop: onstack=%d  first=%d  last=%d  size=%d\n",
      onstack,first,last,size);
    if (width == sizeof(int)) {   /* assume ints */
      printf("ssort(): top intlist=");
      vec_print_int((int *) base, nel);
    }
#endif  /*SSORT_TEST_SUITE*/

    /*  Handle small cases quickly.
    */
    if (size <= SMALL) {
      if (size < 4) {
        if (size < 2) { }
        else if (size == 2) {
          if (S_LESS(last,first))
            S_SWAP(first,last);
        }
        else if (size == 3) {
          if (S_LESS(first+1,first))
            S_SWAP(first+1,first);
          if (S_LESS(last,first+1))
            S_SWAP(last,first+1);
          if (S_LESS(first+1,first))
            S_SWAP(first+1,first);
        }
      }
      else {  /* 4 <= size <= SMALL */
          /*  Sort the first four elements using direct comparison.
          */
          if (S_LESS(first+3,first))
            S_SWAP(first+3,first);
          if (S_LESS(first+2,first+1))
            S_SWAP(first+2,first+1);

          if (S_LESS(first+1,first))
            S_SWAP(first+1,first);
          if (S_LESS(first+3,first+2))
            S_SWAP(first+3,first+2);

          if (S_LESS(first+2,first+1))
            S_SWAP(first+2,first+1);

          /*  Sort any remaining elements using insertion sort.
          */
          if (size > 4) {
            char *fbase;        /* first element in current area */

            fbase = cbase+width*first;

            memcpy(swapbuff,fbase+4*width,width);
            binsert_withdups(swapbuff,fbase,4,width,compar);
            if (size > 5) {
             memcpy(swapbuff,fbase+5*width,width);
             binsert_withdups(swapbuff,fbase,5,width,compar);
             if (size > 6) {
              memcpy(swapbuff,fbase+6*width,width);
              binsert_withdups(swapbuff,fbase,6,width,compar);
              if (size > 7) {
               memcpy(swapbuff,fbase+7*width,width);
               binsert_withdups(swapbuff,fbase,7,width,compar);
               if (size > 8) {
                memcpy(swapbuff,fbase+8*width,width);
                binsert_withdups(swapbuff,fbase,8,width,compar);
                if (size > 9) {
                 memcpy(swapbuff,fbase+9*width,width);
                 binsert_withdups(swapbuff,fbase,9,width,compar);
                 if (size > 10) {
                  memcpy(swapbuff,fbase+10*width,width);
                  binsert_withdups(swapbuff,fbase,10,width,compar);
                  if (size > 11) {
                   memcpy(swapbuff,fbase+11*width,width);
                   binsert_withdups(swapbuff,fbase,11,width,compar);
                   if (size > 12) {
                    memcpy(swapbuff,fbase+12*width,width);
                    binsert_withdups(swapbuff,fbase,12,width,compar);
                    if (size > 13) {
                     memcpy(swapbuff,fbase+13*width,width);
                     binsert_withdups(swapbuff,fbase,13,width,compar);
                     if (size > 14) {
                      memcpy(swapbuff,fbase+14*width,width);
                      binsert_withdups(swapbuff,fbase,14,width,compar);
                      if (size > 15) {
                       memcpy(swapbuff,fbase+15*width,width);
                       binsert_withdups(swapbuff,fbase,15,width,compar);
                      }
                     }
                    }
                   }
                  }
                 }
                }
               }
              }
             }
            }
          }
        }

        /*  Pop stack.
        */
        if (onstack) {
          onstack--;
          first = starts[onstack];
          last = ends[onstack];
        }
        else
          break;
    }
    else {
#ifndef RANDOMIZE_SSORT
      /*  Determine partition.  Take 4 elements evenly spaced throughout the
          set, and put them in slots first, first+1, last-1, and last,
          with these four elements sorted in ascending order.  We use elements
          first+1 and last-1 as the elements that define the three partitions.
          Doing this does not help if we have randomized the list to begin
          with.
      */
      stuff

#endif  /* RANDOMIZE_SSORT */
      if (needtosortfirst8) {
        /*  Grab 8 random elements from throughout the partition, and put
            them into first .. first+7.  If we did not do this, then if we
            are in a left (or right) partition, the left (right) 3 elements
            are smaller than usual, so we are not getting a good sample
            for our median selectors.
        */
        for (i = 0; i < 8; i++) {
          j = first + (random() % (size-i));
          ASSERT(j <= last);
          S_SWAP(first + i, j);
        }

        /*  Note: it is slight overkill to completely sort these, because
            elments 0 .. 2 do not have to be sorted amongst themselves, nor
            do elements 5 .. 7.  This could be taken advantage of somehow,
            probably.
        */
        starts[onstack] = first;
        ends[onstack++] = last;
        last = first+7;
        needtosortfirst8 = 0;
        continue;
      }
      needtosortfirst8 = 1;

      /*  Elements we want are in first, ..., first+7.
      */


      /*  The partition-defining elements are in first+1 and last-1.  Make a
          linear pass through all elements (except first, first+1, last-1,
          and last-2), and note which partition each element goes in.
          Also count how many elements will be put into each partition.

          Uses howmany0 and howmany2 as a heuristic deciding for which
          comparison sequence is most likely to require only one comparison.
          This assumes that past comparisons are a good predictor of future
          comparisons.
      */
      howmany0 = 3;
      howmany1 = 1;
      howmany2 = 4;
      compare1 = first+3;
      compare2 = first+4;
#ifdef  SSORT_TEST_SUITE
      if (width == sizeof(int)) {
        printf("ssort: partition values are %6d and %6d\n",
          * (int *) (cbase+compare1*width), * (int *) (cbase+compare2*width));
      }
#endif  /*SSORT_TEST_SUITE*/
      for (i = first + 8; i <= last; i++) {
        if (howmany2 > howmany0) {
          if (S_LESS(compare2,i)) {
            newloc[i] = 2;
            howmany2++;
          }
          else if (S_LESS(compare1,i)) {
            newloc[i] = 1;
            howmany1++;
          }
          else {
            newloc[i] = 0;
            howmany0++;
          }
        }
        else {
          if (S_LESS(i,compare1)) {
            newloc[i] = 0;
            howmany0++;
          }
          else if (S_LESS(i, compare2)) {
            newloc[i] = 1;
            howmany1++;
          }
          else  {
            newloc[i] = 2;
            howmany2++;
          }
        }
#ifdef  ODEBUG
        printf("list[%d]=%d: put into partition %d\n",
          i,*(int*)(cbase+i*sizeof(int)),newloc[i]);
#endif  /*ODEBUG*/
      }
      /*  Determine partition boundaries.
      */
      fill0 = first+3;
      right0 = first+howmany0-1;
      fill1 = right0+1;
      right1 = right0+howmany1;
      fill2 = right1+1;
      right2 = last-3;

      /*  Put partition-defining elements in place.
      */
      /* Partition 2 */
      S_PSWAP(first+7,last);
      S_PSWAP(first+6,last-1);
      S_PSWAP(first+5,last-2);
      S_PSWAP(first+4,fill2);

      fill2++;

      /* Partition 1 */
      S_PSWAP(first + 3, fill1);
      fill1++;


#ifdef  SSORT_TEST_SUITE
      printf("after partition element swaps:\n");
      printf("howmany0=%6d   fill0=%6d   right0=%6d\n", howmany0,fill0,right0);
      printf("howmany1=%6d   fill1=%6d   right1=%6d\n", howmany1,fill1,right1);
      printf("howmany2=%6d   fill2=%6d   right2=%6d\n",howmany2,fill2,right2);
      vec_print_int((int*)cbase,nel);
#endif  /*SSORT_TEST_SUITE*/

      /*  Put other elements in place. 

          Note:  this algorithm for putting elements in place may or may not
          have a problem with preserving randomness.  This has not been
          investigated.  In fact, several subtelties haven't been looked at,
          including what duplicate elements does to the execution time.
      */
      while (1) {
        /*  Find an out of place element in partition 0 (if there is one).
        */
        while ((fill0 <= right0) && (newloc[fill0] == 0))
          fill0++;
        /*  Find an out of place element in partition 1 (if there is one).
        */
        while ((fill1 <= right1) && (newloc[fill1] == 1))
          fill1++;
        if ((fill0 > right0) && (fill1 > right1))
          break;
        /*  See if we can swap out of place elements in partitions 0 and 1.
        */
        if ((fill0 <= right0) && (newloc[fill0] == 1) &&
            (fill1 <= right1) && (newloc[fill1] == 0)) {
          S_SWAP(fill0,fill1);
          fill0++;
          fill1++;
        }
        else {
          /*  At this point, at least one of the following is true:
                (1)   fill0 points to an out of place element in partition 0
                      that belongs in partition 2.
                (2)   fill1 points to an out of place element in partition 1
                      that belongs in partition 2.
              Find an out of place element in partition 2 and swap it the
              appropriate out of place element from partition 0 or 1.
          */
          ASSERT(fill2 <= right2);
          while (newloc[fill2] == 2) {
            fill2++;
            ASSERT(fill2 <= right2);
          }
          if (newloc[fill2] == 0) {
            /* only possibilities are 1 2 0, 2 0 0, and 2 2 0 */
            ASSERT(fill0 <= right0);
            S_SWAP(fill0,fill2);
            if (newloc[fill0] != 2) {
              ASSERT(newloc[fill0] == 1);
              ASSERT(fill1 <= right1);
              ASSERT(newloc[fill1] == 2);
              S_SWAP(fill1,fill2);
              fill1++;
            }
            fill0++;
          }
          else {
            /* only possibilities are 1 2 1, 2 0 1, and 2 2 1 */
            ASSERT(newloc[fill2] == 1);
            ASSERT(fill1 <= right1);
            S_SWAP(fill1,fill2);
            if (newloc[fill1] != 2) {
              ASSERT(newloc[fill1] == 0);
              ASSERT(fill0 <= right0);
              ASSERT(newloc[fill0] == 2);
              S_SWAP(fill0,fill2);
              fill0++;
            }
            fill1++;
          }
          fill2++;
        }
      }
      ASSERT(fill0 == right0+1);
      ASSERT(fill1 == right1+1);

      /*  Place the partitions on the stack.  We could do this in an order
          designed to minimize the stack size, but this would gain us
          nothing since we allocate memory for things other than
          the stack.  Do not put partitions onto the stack if they
          contain less than 2 elements.

          We do this before actually sorting the partitions at this level
          to save some assignments.
      */
      if (right1 > right0+1) {
        starts[onstack] = right0+1;
        ends[onstack++] = right1;
      }
      if (last > right1+1) {
        starts[onstack] = right1+1;
        ends[onstack++] = last;
      }

      /* first is already set correctly */
      last = right0;
    }
  }
}

/*  Insertion sort.
*/
void
isort(
void            *base,
const size_t    nel,
const size_t    width,
int   (*compar) P((const void *, const void *))
)
{
  char  *cbase = base;
  int   i;
  char  *swapbuff = (char *) alloca(width);

#ifdef  ISORT_COUNT
  isort_moves = 0;
#endif  /*ISORT_COUNT*/
  for (i = 1; i < nel; i++) {
    memcpy(swapbuff,cbase+i*width,width);
    binsert_withdups(swapbuff,cbase,i,width,compar);
  }
}
